#ifndef PERLHOOKS_HEADER
#define PERLHOOKS_HEADER

void start_vote(int cid, const char* chatmsg, const char* desc, void* command);
int vote_active();
void vote_enforce(int result);
void do_send_chat(int chatter_cid, int target, const char* text);
void send_broadcast(const char* text, int cid);
void restart(int warmup);
void game_pause(int state);
void endround();
void set_team(int cid, int team);
int client_team(int cid);
void client_statistics(int cid, int* shots, int* hits, int* kills, int* deaths);

#endif
